/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Abstrakt.AbstrStrom;

import Abstrakt.IIterator;
import Abstrakt.ADL.*;

/**
 * Abstraktní datová struktura - k cestný neuspořádaný strom
 * @author Administrator
 */
public abstract class AbstrStrom<objType> implements IAbstrStrom<objType> {

    /**
     * Vnitřní trída zapouzdřující jednotlivé uzly a listy AbstrStromu
     * @param <objTypeItem> datová struktura ukládaných dat
     */
    private class Item<objTypeItem> {

        /**
         * datový kontejner (struktura) uchovávající schranka ve struktuře
         */
        private objTypeItem data;
        /**
         * Rodič Prvku - pokud je null znamená to, že je to kořen stromu
         */
        private Item<objTypeItem> parent;
        /**
         * ADL synů, pokud je ADL null zmanená to, že je Prvek listem (koncovým
         * uzlem) a ne vnitřním uzlem
         */
        private IAbstrDoubleList<Item<objTypeItem>> iAbstrDoubleList = new AbstrDoubleList<Item<objTypeItem>>() {
        };

        /**
         * bezparametrický konstruktor
         */
        public Item() {
        }

        /**
         * Parametrický konstruktor
         * @param schranka Datový kontejner nesoucí vkládaná schranka (strukturu)
         */
        public Item(objTypeItem data) {
            this.data = data;
        }

        /**
         * Parametrický konstruktor
         * @param data datový kontejner
         * @param parent ukazatel na rodiče prvku
         */
        public Item(objTypeItem data, Item<objTypeItem> parent) {
            this.data = data;
            this.parent = parent;
        }
    }
    /**
     * Kořen struktury
     */
    private Item<objType> koren;
    /**
     * Aktuální pozice prohlídky AbstrStromu
     */
    private Item<objType> aktualni;

    /**
     * bezparametriský konstruktor
     */
    public AbstrStrom() {
    }

    /**
     * veřejný parametrický konstruktor, který po zavolání nastaví kořen a
     * zároveň aktuální prvek nastaví na kořen
     * @param koren
     */
    public AbstrStrom(objType koren) {
        vytvor();
        this.koren.data = koren;
    }

    /**
     * privátní parametrický konstruktor, který po zavolání nastaví kořen a
     * zároveň aktuální prvek nastaví na kořen
     * @param koren
     */
    private AbstrStrom(Item<objType> koren) {
        vytvor();
        this.koren = koren;
        aktualniJeKoren();
    }

    /**
     * Iterátor procházející celý abstrstrom
     * @param <objIt> datový kontejner
     */
    private class Iterator<objIt> implements IIterator<objIt> {

        /**
         * Typ prohlídky AbstrStromu (DO_HLOUBKY/DO_SIRKY)
         */
        TypProhlidky typProhlidky;
        /**
         * Aktuální pozice při procházení AbstrStromu (obdoba položky this.aktualni)
         */
        private Item<objType> position = new Item<objType>();
        /**
         * Vnitřní pomocná struktora (fronta/zásobník) dle typu procházení struktury
         */
        private IProhlidka<Item<objType>> prohlidka;
        /**
         * Obsahuje datové kontejnery, které jsou již připraveny pro navrácení
         */
        private IProhlidka<objType> prohlidka_data = new Abstrakt.AbstrStrom.DoSirky.Prohlidka<objType>();

        /**
         * Parametrický konstruktor iterátoru
         * @param typProhlidky (DO_HLOUBKY/DO_SIRKY)
         */
        public Iterator(TypProhlidky typProhlidky) {
            this.typProhlidky = typProhlidky;
            switch (typProhlidky) {
                case DO_HLOUBKY:
                    this.prohlidka = new Abstrakt.AbstrStrom.DoHloubky.Prohlidka<Item<objType>>();
                    break;
                case DO_SIRKY:
                    this.prohlidka = new Abstrakt.AbstrStrom.DoSirky.Prohlidka<Item<objType>>();
                    break;
            }
            this.naplnProhlidku();
        }

        /**
         * Prochází AbstrStrom a ukládá data do fronty/zásobníku pro výpis
         */
        private void naplnProhlidku() {
            if (jePrazdny()) {
                return;
            }

            if (!this.prohlidka.jePrazdny()) {
                return;
            }

            zpristupniKoren();
            this.prohlidka.set(aktualni);
            this.prohlidka_data.set(aktualni.data);

            while (!this.prohlidka.jePrazdny()) {
                RekurzeProhlidka();
            }
        }

        /**
         * Volá sama sebe, při porcházení DO_HLOUBKY
         */
        private void RekurzeProhlidka() {
            this.position = this.prohlidka.get();

            IIterator<Item<objType>> it;
            it = this.position.iAbstrDoubleList.vytvorIterator();

            while (it.hasNext()) {
                Item<objType> uzel = it.next();

                this.prohlidka_data.set(uzel.data);
                if (!uzel.iAbstrDoubleList.jePrazdny()) { // je uzlem
                    this.prohlidka.set(uzel);
                    if (typProhlidky == TypProhlidky.DO_HLOUBKY) {
                        RekurzeProhlidka();
                    }
                }
            }
        }

        /**
         * Returns <tt>true</tt> if the iteration has more elements. (In other
         * words, returns <tt>true</tt> if <tt>next</tt> would return an element
         * rather than throwing an exception.)
         *
         * @return <tt>true</tt> if the iterator has more elements.
         */
        public boolean hasNext() {
            return !(this.prohlidka_data.jePrazdny());
        }

        /**
         * Returns the next element in the iteration.
         *
         * @return the next element in the iteration.
         * @exception NoSuchElementException iteration has no more elements.
         */
        public objIt next() {
            if (jePrazdny()) {
                return null;
            }
            return (objIt) this.prohlidka_data.get();
        }

        /**
         * TO samé, co netx();
         * @return the next element in the iteration.
         */
        public objIt dalsi() {
            return next();
        }

        /**
         * To samé, jako hasNext();
         * @return <tt>true</tt> if the iterator has more elements.
         */
        public boolean jeDalsi() {
            return hasNext();
        }
    }

    /**
     * Vytvoří novou instanci kořene a nastaví kořen na aktuální
     */
    public void vytvor() {
        this.koren = new Item<objType>();
        aktualniJeKoren();
    }

    /**
     * nastaví kořen na aktuální
     */
    private void aktualniJeKoren() {
        this.aktualni = this.koren;
    }

    /**
     * Zrušení celého seznamu
     */
    public void zrus() {
        if (jePrazdny()) {
            return;
        }
        this.koren.data = null;
        this.koren.iAbstrDoubleList.Zrus();
        this.koren.iAbstrDoubleList = null;
        this.koren.parent = null;
        this.koren = null;

        this.aktualni.data = null;
        this.aktualni.iAbstrDoubleList = null;
        this.aktualni.parent = null;
        this.aktualni = null;
    }

    /**
     * Je prázdný AbstrStrom?
     * @return true = je prázdný, false = není prázdný
     */
    public boolean jePrazdny() {
        return this.koren == null;
    }

    /**
     * Vloží libovolný prvek do AbstrStromu, metoda sama rozpozná o jaký prvek
     * se jedná
     * XXX: pokud není zavolané vytvoř, jak se zachovat? Výjimka, zavolat vytvoř, nebo return bez vložení?
     * @param data datový kontejner
     */
    private void vlozPrvek(objType data) {
        // Vlož kořenový
        if (this.koren == null) {
            this.koren = new Item<objType>(data, null);
            this.aktualni = this.koren;
            return;
        }

        // Vlož jakýkoliv jiný prvek
        Item<objType> item = new Item<objType>(data);
        item.parent = this.aktualni;
        this.aktualni.iAbstrDoubleList.vlozPosledni(item);
    }

    /**
     * Vloží kořen - zavolá funkci pro vkládání prvku, která si již obhospodaří,
     * jaký prvek je vkládaný, zda-li se jedná o kořen, či list
     * @param data datový kontejner
     */
    public void vlozKoren(objType data) {
        zrus();
        vlozPrvek(data);
    }

    /**
     * Vloží list - zavolá funkci pro vkládání prvku, která si již obhospodaří,
     * jaký prvek je vkládaný, zda-li se jedná o kořen, či list
     * @param data datový kontejner
     */
    public void vlozList(objType data) {
        vlozPrvek(data);
    }

    /**
     * Odebere podstrom, který se nachází pod aktuálním ukazatelem
     * @return IAbstrStrom - celá nová struktura
     */
    public IAbstrStrom<objType> odeberPodstrom() {
        if (jePrazdny()) {
            return null;
        }

        IAbstrStrom<objType> abstrStrom = new AbstrStrom<objType>(this.aktualni) {
        };
        this.aktualni.parent.iAbstrDoubleList.odeberAktualni();
        this.aktualni.parent = null;
        this.aktualni = this.koren;

        // dodělané dodatečně, ale mělo by fungovat :)
        if (this.aktualni == this.koren) {
            zrus();
        }
        
        return abstrStrom;
    }

    /**
     * Odebere list ze seznamu
     * TODO: dodělat ošetření, pokud je odebíraný prvek již poslední ve stromu (je kořenem)
     * @return datový kontejner
     */
    public objType odeberList(int index) {
        if (jePrazdny()) {
            return null;
        }

        IIterator<Item<objType>> iIterator = aktualni.iAbstrDoubleList.vytvorIterator();
        Item<objType> item = new Item<objType>();

        int i = 1;
        while (iIterator.hasNext()) {
            item = iIterator.next();
            if (i == index) {
                if (item.iAbstrDoubleList.jePrazdny()) {
                    aktualni.iAbstrDoubleList.odeberAktualni();
                    return item.data;
                }
                return null;
            }
            i++;
        }
        return null;
    }

    /**
     * Zpřístupní kořen, nastaví kořen na aktuální
     * @return datový kontejner
     */
    public objType zpristupniKoren() {
        if (jePrazdny()) {
            return null;
        }
        aktualniJeKoren();
        return this.koren.data;
    }

    /**
     * Mělo by být OK
     * @return  datový kontejner
     */
    public objType zpristupniSyna(int index) {
        if (jePrazdny()) {
            return null;
        }

        return najdiSyna(index, true);
    }

    /**
     * 
     * @param typProhlidky true = do šířky, false = do hloubky
     * @return Iterátor
     */
    public IIterator<objType> vytvorIterator(TypProhlidky typProhlidky) {
        IIterator<objType> iterator = new Iterator(typProhlidky);
        return iterator;
    }

    /**
     * Vrátí počet prvků, které obsahuje celý ABstrStrom
     * @return počet prvků AbstrStromu
     */
    public int dejMohutnost() {
        if (jePrazdny()) {
            return 0;
        }
        IIterator<objType> iIterator = new Iterator<objType>(TypProhlidky.DO_SIRKY);
        int cetnost = 0;
        while (iIterator.hasNext()) {
            cetnost++;
            iIterator.next();
        }
        return cetnost;
    }

    /**
     * Dej kořen vrátí datový kontejner kořene
     * @return datový kontejner
     */
    public objType dejKoren() {
        if (jePrazdny()) {
            return null;
        }
        return this.koren.data;
    }

    /**
     * Vrátí datový kontejner specifického syna
     * @param index číselné vyjádření prvku mezi syny
     * @return datový kontejner
     */
    public objType dejSyna(int index) {
        if (jePrazdny()) {
            return null;
        }
        return najdiSyna(index, false);
    }

    /**
     * Nalezení potomka pro dejSyna a zpristupniSyna
     * @param index Index syna
     * @param setaktualni boolean: true - nastaví syna aktuální
     * @return datový kontejenr
     */
    private objType najdiSyna(int index, boolean setaktualni) {
        IIterator<Item<objType>> iIterator = aktualni.iAbstrDoubleList.vytvorIterator();
        Item<objType> item = new Item<objType>();

        int i = 1;
        while (iIterator.hasNext()) {
            item = iIterator.next();
            if (i == index) {
                if (setaktualni) {
                    this.aktualni = item;
                }
                return item.data;
            }
            i++;
        }
        return null;
    }
}
