/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Abstrakt.PriorFronta;

import Abstrakt.IIterator;
import Abstrakt.ADL.AbstrDoubleList;

/**
 * Abstraktní datová struktura Prioritní Fronta postavená na struktuře ADL
 * @author Administrator
 */
public class PriorFronta<T, U> {

    /**
     * Vnitřní třída zapouzdřující datový kontejner a prioritu podle které se řadí
     * @param <T> datová struktura
     * @param <U> priorita pro třídění
     */
    private class Item<T, U> {

        T data;
        U priorita;

        public Item() {
        }

        public Item(T data, U priorita) {
            this.data = data;
            this.priorita = priorita;
        }
    }
    /**
     * Vnitřní uspořádání PriorFronty
     */
    private AbstrDoubleList<Item<T, U>> abstrDoubleList;

    /**
     * Iterátor prioritní fronty, musí zde být implementován kvůli vnitřní třídě
     */
    private class Iterator implements IIterator<T> {

        private IIterator<Item<T, U>> iIterator = abstrDoubleList.vytvorIterator();

        public T dalsi() {
            return next();
        }

        public boolean hasNext() {
            return this.iIterator.hasNext();
        }

        public boolean jeDalsi() {
            return hasNext();
        }

        public T next() {
            return this.iIterator.next().data;
        }
    }

    /**
     * bezparametrický konsturktor, který pouze alokuje paměť pro ADL
     */
    public PriorFronta() {
        this.abstrDoubleList = new AbstrDoubleList<Item<T, U>>() {
        };
    }

    /**
     * Obdoba bezpamaretrického konstruktoru, pouze se nikdy neprovede vnitřek,
     * jelikož je vždy předem volán konstruktor
     */
    public void vytvor() {
        if (jePrazdna()) {
            this.abstrDoubleList = new AbstrDoubleList<Item<T, U>>() {
            };
        }
    }

    /**
     * Zrušení prioritní fronty
     */
    public void zrus() {
        if (!jePrazdna()) {
            this.abstrDoubleList.Zrus();
        }
    }

    /**
     * Zjištění, zdali je PrioritníFronta prázdná
     * @return true = je prázdná, false = není prázdná
     */
    public boolean jePrazdna() {
        return this.abstrDoubleList.jePrazdny();
    }

    /**
     * Vrací počet prvků uloženým v Prioritní Frontě
     * @return počet prvků
     */
    public int mohutnost() {
        if (jePrazdna()) {
            return 0;
        }
        IIterator<Item<T, U>> iIterator = vytvorIterator();
        int cetnost = 0;
        while (iIterator.hasNext()) {
            cetnost++;
            iIterator.next();
        }
        return cetnost;
    }

    /**
     * Vloží prvek do Prioritní Fronty
     * @param data datový kontejner
     * @param jednotka priorita, dle které se bude třídit
     * @param priority komparátor s prioritou
     */
    public void vloz(T data, U jednotka, Comparable priority) {
        Item<T, U> item = new Item<T, U>(data, jednotka);

        IIterator<Item<T, U>> iIterator = this.abstrDoubleList.vytvorIterator();
        while (iIterator.hasNext()) {
            Item<T, U> aktualni = iIterator.next();

            if (priority.compareTo(aktualni.priorita) >= 0) { // THX Belzebub za vysvětlení, jak použít compareTo
                this.abstrDoubleList.vlozPredchudce(item);
                return;
            }
        }
        this.abstrDoubleList.vlozPosledni(item);
    }

    /**
     * Odebere poslední (maximální) prvek v Prioritní frontě
     * @return datový kontejner
     */
    public T odeberMax() {
        if (jePrazdna()) {
            return null;
        }
        return this.abstrDoubleList.odeberPosledni().data;
    }

    /**
     * Zpřístupní data z posledního (maximáního) prvku v Prioritní forntě
     * @return datový kontejner
     */
    public T zpristupniMax() {
        if (jePrazdna()) {
            return null;
        }
        return this.abstrDoubleList.zpristupniPosledni().data;
    }

    /**
     * Vyhledá zda-li se nějaký prvek v Prioritní frontě rovná hledanému prvku
     * @param priorita prvek, který požadujeme nalézt
     * @return datový kontejner
     */
    public T vyhledej(U priorita) {
        if (jePrazdna()) {
            return null;
        }
        IIterator<Item<T, U>> iIterator = this.abstrDoubleList.vytvorIterator();
        while (iIterator.hasNext()) {
            Item<T, U> item = iIterator.next();
            if (item.priorita.equals(priorita)) {
                return item.data;
            }
        }
        return null;
    }

    /**
     * Vytvoří iterátor
     * @return iterátor Prioritní fronty
     */
    public IIterator vytvorIterator() {
        return new Iterator();
    }

    /**
     * Převede datovou složku na textový řetězec
     * @return textový řetězec z datového kontejneru
     */
    @Override
    public String toString() {
        String result = null;
        IIterator it = abstrDoubleList.vytvorIterator();
        while (it.hasNext()) {
            result += it.next().toString() + "\r\n";
        }
        return result;
    }
}
